// Copyright 2003 Regents of the University of California

// Astropulse is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// Astropulse is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with Astropulse; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* ap_fold.C */
/* function for folding */

#ifndef _WIN32
#include "config.h"
#include <unistd.h>
#undef BOINC_APP_GRAPHICS
#else
#include <windows.h>
#define BOINC_APP_GRAPHICS 1
#endif

#include <cassert>
#include <cstdio>
#include <cstdlib>

#include "boinc_api.h"
#include "util.h"

#include "astropulse.h"

#include <cmath>

#define ITMAX 1000
#define EPS 1e-6
#define FP_EPS 1.0e-35

float lcgf(float a, float x);
static float gammln(float a);

namespace Astropulse {

    static const int pnum[]={ 137, 139, 149, 151, 157, 163, 167, 173, 179, 181, 191, 193, 197,
                              199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 269, 271,
                              277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 347, 349, 353,
                              359, 367, 373, 379, 383, 389, 397, 401, 409, 419, 421, 431, 433,
                              439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503, 509,
                              521, 523, 541, 547, 557, 563, 569, 571, 577, 587, 593, 599, 601,
                              607, 613, 617, 619, 631, 641, 643, 647, 653, 659, 661, 673, 677,
                              683, 691, 701, 709, 719, 727, 733, 739, 743, 751, 757, 761, 769,
                              773, 787, 797, 809, 811, 821, 823, 827, 829, 839, 853, 857, 859,
                              863, 877, 881, 883, 887, 907, 911, 919, 929, 937, 941, 947, 953,
                              967, 971, 977, 983, 991, 997, 1009 };

    /* General plan: use the prime numbers as frequencies, so we don't do
       repetitive folds.  For each base frequency, do that fold, then 2x, then
       2x that, etc ... step periods by a factor of dp each time  */

    /* This version doesn't return any results yet */

    int Science::ffa(float *data, float dp) {
        int n_bins = client.state.fold_buf_size;
        float period, freq;
        float subfreq;
        /* Max freq for folding - period of two samples */
        float maxfreq=(float)(client.state.datasize*4.0/2.0);
        int per_int;
        int i,j,k,l,m;
        static float *temp = NULL;
        int n_per;
        int n_thresh = client.state.dm_chunk*1;
	ap_signal s;

	FILE* writefile = fopen("christest.dat", "a+");
	
        if (temp == NULL) {
            temp = (float *)malloc(sizeof(float) * n_bins);
        }

        client.status= "Computing fast fold.";
        /* Loop over all periods - two for's */
        for (i=0; i<136; i++) {
            for (freq=(float)pnum[i]; freq<pnum[i+1]; freq*=(1+dp)) {
                period = (float)n_bins/freq;
                /* Do initial fold */
                per_int = (int)ceil(period);
                n_per = (int)freq;
                memset(temp, 0, sizeof(temp));
                for (j=0; j<per_int; j++) {
                    for (k=0; k<n_per; k++) {
                        l = (int)(period*(float)k + 0.5);
                        temp[j] += data[l];
                    }
                }
                n_thresh = client.state.dm_chunk*n_per; // co-add level for threshold
                /* Look for spikes here  - return results */
                /* Need to figure out thresh */
                  for (m=0; m<per_int; m++) {
		    // if probability of existence > .000001, valid data
		    if (lcgf((float) n_thresh, temp[m]) > -6+log((float)per_int)) { 
		      time_series_b_t time_series;
		      time_series.build(temp, m, per_int);
		      
		      client.state.pulse_jd = client.state.jd0 + (client.state.data_chunk_now + client.state.dm_chunk)
			/(client.state.datasize)*(client.state.jd1-client.state.jd0);
			      s.time=client.state.pulse_jd;
			      s.peak_bin = client.state.data_chunk_now*4 + m;
			      s.peak_power = temp[m];
			      s.scale = client.state.fold_level;
			      s.dm = client.state.dm_sign*client.state.dm_chunk_now;
			      s.period = period;
			      s.time_series=time_series.data;
			      client.outfile.output_result(s);
		    }
		    
		  }
	    }
		  /* Now do powers of two */
		  for (subfreq=freq; subfreq<maxfreq; subfreq*=2) {
		    period = (float)n_bins/subfreq;
		    per_int = per_int >> 1;
		    n_thresh = n_thresh << 1;
		      for (j=0; j<per_int; j++) {
                        temp[j] = temp[j] + temp[j+per_int];
		      }
		      /* Look for spikes, return results */
		      
		      for (m=0; m<per_int; m++) {
			if(lcgf((float) n_thresh, temp[m]) > -6+log((float)per_int)) {
			    time_series_b_t time_series;
			    time_series.build(temp, m, per_int);
			    
			    client.state.pulse_jd = client.state.jd0 + (client.state.data_chunk_now + client.state.dm_now/client.state.dm_chunk)
			/(client.state.datasize)*(client.state.jd1-client.state.jd0);
			      s.time=client.state.pulse_jd;
			      s.peak_bin = client.state.data_chunk_now*4 + m;
			      s.peak_power = temp[m];
			      s.scale = client.state.fold_level;
			      s.dm = client.state.dm_sign*client.state.dm_chunk_now;
			      s.period = period;
			      s.time_series=time_series.data;
			      client.outfile.output_result(s);

			 }
		      
		      }
		  }
	}
	
        return(0);
    }	
}

/*  Log of the complete gamma function  */
static float gammln(float a) {
  double x,y,tmp,ser;
  static double cof[6]={76.18009172947146,-86.50532032941677,
                        24.01409824083091,-1.231739572450155,
                        0.1208650973866179e-2,-0.5395239384953e-5};
  int j;

  y=x=a;
  tmp=x+5.5;
  tmp -= (x+0.5)*log(tmp);
  ser=1.000000000190015;
  for (j=0;j<=5;j++) ser += cof[j]/++y;
  return (float)(-tmp+log(2.5066282746310005*ser/x));
}

/* Log of the compliment of the incomplete gamma function
 * log(1-P(a,x)) valid only for a<(x+1)
 */
float lcgf(float a, float x) {
  int i;
  float an,b,c,d,del,h,gln=gammln(a);

  //  BOINCASSERT(x>=(a+1));
  b=x+1.0f+a;
  c=1.0f/(float)FP_EPS;
  d=1.0f/b;
  h=d;
  for (i=1;i<=ITMAX;i++) {
    an = -i*(i-a);
    b += 2.0f;
    d=an*d+b;
    if (fabs(d)<FP_EPS) d=(float)FP_EPS;
    c=b+an/c;
    if (fabs(c)<FP_EPS) c=(float)FP_EPS;
    d=1.0f/d;
    del=d*c;
    h*=del;
    if (fabs(del-1.0)<EPS) break;
  }
  //BOINCASSERT(i<ITMAX);
  return (float)(log(h)-x+a*log(x)-gln);
}
